/*
 * This file is part of SPORE.
 *
 * Copyright (C) 2016, the SPORE team (see AUTHORS).
 *
 * SPORE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * SPORE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SPORE.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * File:   reward_in_proxy.h
 *
 * This file is based on music_cont_in_proxy.h which is part of NEST
 * (Copyright (C) 2004 The NEST Initiative).
 * See: http://nest-initiative.org/
 */

#ifndef REWARD_IN_PROXY_H
#define REWARD_IN_PROXY_H

#include "nest.h"
#include "event.h"
#include "ring_buffer.h"
#include "spikecounter.h"
#include "connection.h"

#include "namedatum.h"
#include "universal_data_logger.h"

#include <music.hh>

#include "tracing_node.h"

namespace spore
{

/**
 * @brief A device allows to receive reward traces from another process.
 * 
 * This device allows to receive a real-valued piecewise constant trace
 * from another process using the MUSIC interprocess communication API.
 * A typical application for this node is to receive a reward signal and
 * communicate it to synapses that are trained using reward-modulated
 * plasticity. The number of traces that are received by this device is given
 * by the width of the MUSIC port. The received traces can be read out using
 * the TracingNode::get_trace() member function.
 * 
 * @see SynapticSamplingRewardGradientConnection
 */
class RewardInProxy : public TracingNode
{
public:

    RewardInProxy();

    bool has_proxies() const
    {
        return false;
    }

    bool one_node_per_process() const
    {
        return true;
    }

    virtual void get_status(DictionaryDatum& d) const;
    virtual void set_status(const DictionaryDatum& d);

protected:

    virtual void init_buffers_();
    virtual void init_state_(const Node&);

    virtual void calibrate();

    virtual void update(nest::Time const&, const long, const long);

    MUSIC::ContInputPort* reward_in_; //!< The MUSIC cont port for reward input
    std::vector< double > reward_in_buffer_; //!< The mapped buffer for receiving the reward values via MPI

    // ------------------------------------------------------------

    /**
     * @brief Class holding state variables of the proxy.
     */
    struct State_
    {
        State_(); //!< Sets default state value

        void get(DictionaryDatum&) const; //!< Store current values in dictionary

        bool published_; //!< indicates whether this node has been published already with MUSIC
        int port_width_; //!< the width of the MUSIC port
    };

    /**
     * @brief Class holding parameter variables of the proxy.
     */
    struct Parameters_
    {
        Parameters_(); //!< Sets default parameter values

        void get(DictionaryDatum&) const; //!< Store current values in dictionary
        void set(const DictionaryDatum&, State_&); //!< Set values from dicitonary

        std::string port_name_; //!< the name of MUSIC port to read from
        float delay_; //!< the accepted delay for the MUSIC connection
    };

    State_ S_;
    Parameters_ P_;
};

}

#endif
